/*******************************************************************************
 * Copyright 2007  Hon Hai Precision Ind. Co. Ltd.
 * All Rights Reserved.
 * No portions of this material shall be reproduced in any form without the
 * written permission of Hon Hai Precision Ind. Co. Ltd.
 *
 * All information contained in this document is Hon Hai Precision Ind.
 * Co. Ltd. company private, proprietary, and trade secret property and
 * are protected by international intellectual property laws and treaties.
 *
 ******************************************************************************/

/******************************************************************************
***    Filename: br_dhcp_filter.c
***    Description:
***       Per NETGEAR repeater spec, dhcp client and dhcp server packets (from
***    stack) need to be filtered in LAN and WAN interfaces respectively.
***    History:
***
*** Modify Reason       Author      Date          Search Flag(Option)
*** ----------------------------------------------------------------------------
*** File Creation       wklin       05/13/2010
*** Port to WNCE3001    pling       07/22/2011
********************************************************************************/
 
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/string.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_bridge.h>

#include "bridge/br_private.h"

//#define LOCAL_OUT_NO_DHCPC(dev) (dev && (dev->acos_flags&BR_LOCAL_OUT_DHCPC_FILTER))
#define LOCAL_OUT_NO_DHCPD(dev) (dev && dev->name && (dev->name[0]=='w'))    /* interface 'wlan0' */

#if 0
static int is_dhcpc_packet(struct sk_buff *skb)
{
    if (skb->mac.ethernet->h_proto == htons(ETH_P_IP)) { /* IP */
        skb->nh.iph = (struct iphdr *)skb->data;
        if (skb->nh.iph->protocol == IPPROTO_UDP) {
            skb->h.uh = (struct udphdr *)(skb->nh.raw + skb->nh.iph->ihl*4);
            if (skb->h.uh->source == htons(68)) { /* dhcp client */
                /*printk("<0>drop localout dhcpc packet @ %s\n",
                        skb->dev?skb->dev->name:"???");
                */
                return 1;
            }
        }
    }

    return 0;
}
#endif

static int is_dhcpd_packet(struct sk_buff *skb)
{
#if 0
    if (skb->mac.ethernet->h_proto == htons(ETH_P_IP)) { /* IP */
        skb->nh.iph = (struct iphdr *)skb->data;
        if (skb->nh.iph->protocol == IPPROTO_UDP) {
            skb->h.uh = (struct udphdr *)(skb->nh.raw + skb->nh.iph->ihl*4);
            if (skb->h.uh->source == htons(67)) { /* dhcp server */
                printk("<0>drop localout dhcpd packet @ %s\n",
                        skb->dev?skb->dev->name:"???");
                return 1;
            }
        }
    }
#endif

    struct ethhdr *eth_hdr = (struct ethhdr *)(skb_mac_header(skb));
    struct iphdr  *ip_hdr;
    struct udphdr *udp_hdr;
    unsigned char *temp;

    if (eth_hdr->h_proto == htons(ETH_P_IP)) { /* IP */
        ip_hdr = (struct iphdr *)skb_network_header(skb);
        if (ip_hdr->protocol == IPPROTO_UDP) {
            temp = (unsigned char *)ip_hdr;
            udp_hdr = (struct udphdr *)&(temp[ip_hdr->ihl*4]);
            if (udp_hdr->source == htons(67)) { /* dhcp server */
#if 0
                printk(KERN_EMERG "drop localout dhcpd packet @ %s\n",
                        skb->dev?skb->dev->name:"???");
#endif
                return 1;
            }
        }
    }

    return 0;
}

static int br_dhcp_filter_hook (
        struct sk_buff *skb,
        unsigned int hook,
        const struct net_device *in,
        const struct net_device *out)
{
    if (LOCAL_OUT_NO_DHCPD(out)) {
        if(is_dhcpd_packet(skb))
            return 1; /* to be dropped */
    }
#if 0
    else if (LOCAL_OUT_NO_DHCPC(out)) {
        if(is_dhcpc_packet(skb))
            return 1; /* to be dropped */
    }
#endif
    return 0; /* accept */
}

static unsigned int br_localout_hook
(
    unsigned int hook,
    struct sk_buff *skb,
    const struct net_device *in,
    const struct net_device *out,
    int (*okfn)(struct sk_buff *)
)
{
    if (br_dhcp_filter_hook(skb, hook, in, out))
        return NF_DROP;
    else
        return NF_ACCEPT;
}


static struct nf_hook_ops br_localout_ops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
    .owner          = THIS_MODULE,
#else
    .list           = {NULL, NULL},
#endif /* AG_LINUX_26 */
    .hook           = br_localout_hook,
    .pf             = PF_BRIDGE,
    .hooknum        = NF_BR_LOCAL_OUT,
    .priority       = INT_MIN,
};

static int br_dhcp_filter_init(void)
{
    struct net_device *dev;
    if (nf_register_hook(&br_localout_ops) < 0) {
        printk("<0>br_dhcp_filter: Unable to register hook @ NF_BR_LOCAL_OUT.\n");
    } 
#if 0
    else {
        printk("<0>NF_BR_LOCAL_OUT registered.\n");
        dev = dev_get_by_name("eth1"); /* client if */
        if (dev) {
            dev->acos_flags |= BR_LOCAL_OUT_DHCPD_FILTER;
            printk("dev=%p %s, acos_flags = %08x\n", dev, dev?dev->name:"???", dev->acos_flags);
        }
        dev = dev_get_by_name("vlan1"); /* eth if */
        if (dev) {
            dev->acos_flags |= BR_LOCAL_OUT_DHCPC_FILTER;
            printk("dev=%p %s, acos_flags = %08x\n", dev, dev?dev->name:"???", dev->acos_flags);
        }
        dev = dev_get_by_name("wl0.1"); /* ap if */
        if (dev) {
            dev->acos_flags |= BR_LOCAL_OUT_DHCPC_FILTER;
            printk("dev=%p %s, acos_flags = %08x\n", dev, dev?dev->name:"???", dev->acos_flags);
        }
    }
#endif
    return 0;
}

static void br_dhcp_filter_exit(void)
{
    nf_unregister_hook(&br_localout_ops);
    printk("<0>NF_BR_LOCAL_OUT removed.\n");
}

module_init(br_dhcp_filter_init);
module_exit(br_dhcp_filter_exit);
