/*******************************************************************************
 * Copyright 2010 Hon Hai Precision Ind. Co. Ltd.
 * All Rights Reserved.
 * No portions of this material shall be reproduced in any form without the
 * written permission of Hon Hai Precision Ind. Co. Ltd.
 *
 * All information contained in this document is Hon Hai Precision Ind.
 * Co. Ltd. company private, proprietary, and trade secret property and
 * are protected by international intellectual property laws and treaties.
 *
 ******************************************************************************/

/******************************************************************************
***    Filename: br_dns_hijack.c
***    Description:
***        1. purpose
***             Kernel module for wireless repeaters.
***             To hijack Domain names:
***                 www.mywifiext.net
***                 www.mywifiext.com
***                 mywifiext.net
***                 mywifiext.com
***             and return the repeaters' own IP address
***             for GUI management.
***
***    History:
***
*** Modify Reason           Author      Date          Search Flag(Option)
*** ----------------------------------------------------------------------------
*** File Creation           Peter Ling  05/19/2010
*** Port to home routers    Peter Ling  03/18/2011
***   for AP mode
********************************************************************************/
 
#include <linux/config.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fcntl.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/string.h>
#include <linux/netfilter_bridge.h>
#include <linux/socket.h>

#include <linux/netfilter_ipv4.h>
#include <net/snmp.h>
#include <net/ip.h>
#include <net/route.h>
#include <net/protocol.h>
#include <net/sock.h>
#include <net/netlink.h>

#include <net/udp.h>
#include <net/raw.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <net/checksum.h>

#include "br_dns_hijack.h"

#define A_TOLOWER(c)        (((c) >= 'A' && (c) <= 'Z') ? ((c)-'A'+'a') : (c))

#if (defined WN3000RPT)
#define WLAN_IFNAME     "wl0.1"
#define LAN_IFNAME      "vlan1"
#define BR_IFNAME       "br0"
#elif (1)//defined WNR3500L)
#define WLAN_IFNAME         "eth1"
#define WLAN_GUEST_IFNAME   "wl0.1"
#define LAN_IFNAME          "eth0"
#define BR_IFNAME           "br0"
#else
#error "WLAN_IFNAME not defined!"
#error "LAN_IFNAME not defined!"
#endif

#define DEFAULT_BR_IP       0xc0a801fa  /* 192.168.1.250 */

#define NIPQUAD(addr) \
        ((unsigned char *)&addr)[0], \
        ((unsigned char *)&addr)[1], \
        ((unsigned char *)&addr)[2], \
        ((unsigned char *)&addr)[3]

/* For hijacking readyshare.routerlogin.net */
#define MAX_READYSHARE_NAME_LEN     64
static char readyshare_name[MAX_READYSHARE_NAME_LEN];
static struct sock *netlink_sock;
static char *readyshare_dev = readyshare_name;
module_param(readyshare_dev, charp, S_IRUGO);
/* foxconn wklin modified start, 05/20/2011, set hijack mode to 1 by default */
static int hijack_mode = 1;
/* foxconn wklin modified end, 05/20/2011 */
static unsigned char *dnsURLForRedirect[] =
{
#if 1/*#ifdef WN3000RPT*/
    "www.mywifiext.net",
    "mywifiext.net",
    "www.mywifiext.com",
    "mywifiext.com",
#else
    "www.routerlogin.net",
    "routerlogin.net",
    "www.routerlogin.com",
    "routerlogin.com",
#endif
};

static void udp_reply(int pid,int seq,void *payload)
{
	struct sk_buff	*skb;
	struct nlmsghdr	*nlh;
	int size=strlen(payload)+1;
	int		len = NLMSG_SPACE(size);
	void		*data;
	int ret;
	
	skb = alloc_skb(len, GFP_ATOMIC);
	if (!skb)
		return;
	nlh= NLMSG_PUT(skb, pid, seq, 0, size);
	nlh->nlmsg_flags = 0;
	data=NLMSG_DATA(nlh);
	memcpy(data, payload, size);
	NETLINK_CB(skb).pid = 0;         /* from kernel */
	NETLINK_CB(skb).dst_group = 0;  /* unicast */
	ret=netlink_unicast(netlink_sock, skb, pid, MSG_DONTWAIT);
	if (ret <0)
	{
		printk("send failed\n");
		return;
	}
	return;
	
nlmsg_failure:			/* Used by NLMSG_PUT */
	if (skb)
		kfree_skb(skb);
}

/* Receive messages from netlink socket. */
static void udp_receive(struct sk_buff  *skb)
{
    u_int	uid, pid, seq, sid;
	void			*data;
	struct nlmsghdr *nlh;
	
	nlh = (struct nlmsghdr *)skb->data;
	pid  = NETLINK_CREDS(skb)->pid;
	uid  = NETLINK_CREDS(skb)->uid;
	sid  = NETLINK_CB(skb).sid;
	seq  = nlh->nlmsg_seq;
	data = NLMSG_DATA(nlh);
	/* printk("recv skb from user space uid:%d pid:%d seq:%d,sid:%d\n",uid,pid,seq,sid); */
    /*
	printk("data is :%s\n",(char *)data);
    */
    /*     
    printk("payload get hijack == %s\n",(char *)(data + 7));
    printk("data %s\n",(char*)data);
    */
    if(strncmp((char*)data,"hijack=1",8) == 0){
        printk("hijack mode == 1\n");
        hijack_mode = 1;
    }else{
        printk("hijack mode == 0\n");
        hijack_mode = 0;
    }
	udp_reply(pid,seq,data);
	return ;
}





static unsigned char *dnsPtrQuery[] = 
{
    "1.0.0.127.in-addr.arpa",
};

int dnsRedirect_getQueryName(unsigned char *buffer, unsigned char *queryName, unsigned int queryNameLen)
{
    unsigned int totalLen=0, byteCount, i;

    for (i=0; ; i++)
    {
        if (*buffer == 0)
        {
            *queryName = '\0';
            break;
        }
        else
        if (i != 0)
            *queryName++ = '.';

        byteCount = *buffer++;
        if (totalLen + byteCount + 1 >= queryNameLen)
            return ERROR;

        totalLen += byteCount + 1;
        memcpy(queryName, buffer, byteCount);
        queryName += byteCount;
        buffer    += byteCount;
    }

    return OK;
}

int stricmp(char *str1, char *str2)
{
    if (!str1 || !str2)
        return 1;

    if (strlen(str1) != strlen(str2))
        return 1;

    for (; *str1; str1++, str2++)
    {
        if ( A_TOLOWER(*str1) != A_TOLOWER(*str2) )
                return 1;
    }

   return 0;
}

int dnsRedirect_getRedirectIndex(unsigned char *queryName)
{   
    int i;


    for (i=0; i<sizeof(dnsURLForRedirect)/sizeof(dnsURLForRedirect[0]); i++)
    {   
        if (stricmp(dnsURLForRedirect[i], queryName) == 0)
            return i;
    }

    /* For hijacking readyshare.routerlogin.com */
    if (strlen(readyshare_name))
    {
#define READYSHARE_SUFFIX1  ".routerlogin.com"
#define READYSHARE_SUFFIX2  ".routerlogin.net"

        char readyshare1[MAX_READYSHARE_NAME_LEN+sizeof(READYSHARE_SUFFIX1)];
        char readyshare2[MAX_READYSHARE_NAME_LEN+sizeof(READYSHARE_SUFFIX2)];
        sprintf(readyshare1, "%s%s", readyshare_name, READYSHARE_SUFFIX1);
        sprintf(readyshare2, "%s%s", readyshare_name, READYSHARE_SUFFIX2);

        if ((stricmp(readyshare1, queryName) == 0) ||
            (stricmp(readyshare2, queryName) == 0))
            return i;
    }

    return ERROR_INDEX;
}

int dnsRedirect_getPtrQueryIndex(unsigned char *queryName)
{
    int i;

    for (i=0; i<sizeof(dnsPtrQuery)/sizeof(dnsPtrQuery[0]); i++)
    {   
        if (stricmp(dnsPtrQuery[i], queryName) == 0)
            return i;
    }

    return ERROR_INDEX;
}

BOOL dnsRedirect_isNeedRedirect(T_IPHeader *pIpHeader, unsigned short *queryType)
{
    unsigned char *pDnsPacket = NULL;
    dnsHeader_T *pDnsPkt = NULL;
    char *queryName; /* wklin added, 05/27/2008 @stack_size */
    unsigned char *c;
    /* wklin added start, 05/27/2008 @stack_size */
    queryName = malloc(C_MAX_DNS_PACKET_SIZE);
    if (!queryName) {
        printf("malloc failed %s:%d\n", __FUNCTION__, __LINE__);
        return FALSE;
    }
    /* wklin added end, 05/27/2008 */
    
    pDnsPkt = (dnsHeader_T *)GetUdpPayload(pIpHeader);
    pDnsPacket = (unsigned char *)pDnsPkt;
    
    BR_DNS_HIJACK_DEBUG( "dnsPkt.qr = %d\n", pDnsPkt->qr);
    if (pDnsPkt->qr == DNS_QUERY)
    {
        unsigned char *pType = NULL;
        unsigned short queryClass = 0;

        /* Note: We assume there is only 1 query in this packet */
        dnsRedirect_getQueryName(pDnsPacket + sizeof(dnsHeader_T), queryName, C_MAX_DNS_PACKET_SIZE);
              
        BR_DNS_HIJACK_DEBUG( "queryName = '%s'\n", queryName) ;
        /* Foxconn modified start, zacker, 04/23/2008, @ipv6_dns_redirect */
        /* Bypass IPv6 DNS query */
        pType = pDnsPacket + sizeof(dnsHeader_T) + strlen(queryName) + 2;
        memcpy(queryType, pType, 2);
        *queryType = ntohs(*queryType);
        memcpy(&queryClass, pType + 2, 2);
        queryClass = ntohs(queryClass);
        
        if (dnsRedirect_getRedirectIndex(queryName) != ERROR_INDEX)
        {
            BR_DNS_HIJACK_DEBUG("DNS redirect found!\n");
            // *queryType = DNS_QUERY_TYPE_IP;
            if ((*queryType == DNS_QUERY_TYPE_IP)||
                (*queryType == DNS_QUERY_TYPE_IPV6)) {
                free(queryName); /* wklin added, 05/27/2008 @stack_size */
            return TRUE;
            }
        }

        /* We need to reply some query to Win OS, otherwise it thinks the DNS server is not responsive 
         * and stop sending queries */
        else
        if (dnsRedirect_getPtrQueryIndex(queryName) != ERROR_INDEX)
        {
            BR_DNS_HIJACK_DEBUG("pointer query\n");
            // *queryType = DNS_QUERY_TYPE_PTR;
            if (*queryType == DNS_QUERY_TYPE_PTR) {
                free(queryName); /* wklin added, 05/27/2008 @stack_size */
            return TRUE;
            }
        }
        /* Foxconn modified end, zacker, 04/23/2008, @ipv6_dns_redirect */ 

        if(hijack_mode == 1){
            free(queryName);
	    return TRUE;
        }
    }
   

    free(queryName); /* wklin added, 05/27/2008 @stack_size */
    return FALSE;
}

UINT16 CalcChecksum(BYTE *pbData, int iLength)
{
    int nLeft = iLength % 2; /* this addr has odd byte? */
    UINT16 wChecksum;
    int iIndex;
    UINT32 dwSum = 0;

    for (iIndex=0; iIndex < iLength-1; iIndex+=2)
    {
        dwSum += (pbData[iIndex] << 8) | (pbData[iIndex+1]);
    }

    if (nLeft != 0)
        dwSum += (pbData[iLength-1] << 8);


    dwSum = (dwSum >> 16) + (dwSum & 0xffff);     /* add hi 16 to low 16 */
    dwSum += (dwSum >> 16);                     /* add carry */
    wChecksum = ~dwSum;                          /* truncate to 16 bits */
    return (wChecksum);

}

UINT16  CalcIpChecksum(T_IPHeader *ptIp)
{
    UINT16 wChecksum;
    AG_UINT16_TO_BUFFER(ptIp->checksum,0);
    wChecksum = CalcChecksum((BYTE *)ptIp, GetIpHeaderLen(ptIp));
    AG_UINT16_TO_BUFFER(ptIp->checksum,htons(wChecksum));
    return wChecksum;
}

UINT16 CalcChecksum2(UINT32 dwPreChecksum, BYTE *pbData2, int iLength2)
{
    int nLeft;
    UINT16 wChecksum;
    int iIndex;


    /* calculate buffer 2*/
    nLeft = iLength2 % 2;
    for (iIndex=0; iIndex < iLength2-1; iIndex+=2)
    {
        dwPreChecksum += (pbData2[iIndex] << 8) | (pbData2[iIndex+1]);
    }

    if (nLeft != 0)
        dwPreChecksum += (pbData2[iLength2-1] << 8);


    dwPreChecksum = (dwPreChecksum >> 16) + (dwPreChecksum & 0xffff);     /* add hi 16 to low 16 */
    dwPreChecksum += (dwPreChecksum >> 16);                     /* add carry */
    wChecksum = ~dwPreChecksum;                          /* truncate to 16 bits */
    return (wChecksum);

}

UINT16 CalcUdpChecksum(T_IPHeader *ptIp)
{
    UINT16 wChecksum;
    T_UDPHeader *ptUdp;
    unsigned int iIpHeaderLen, iIpLen, iUdpLen;
    UINT32 dwPreChecksum=0;

    if (!IsUdpPacket(ptIp))
        return 0;

    iIpHeaderLen = GetIpHeaderLen(ptIp);
    iIpLen = (ptIp->length[0]<<8) + ptIp->length[1];
    iUdpLen = iIpLen - iIpHeaderLen;

    /* point to the tcp packet */
    ptUdp = (T_UDPHeader *)GetIpPayload(ptIp);

    /* clear the original checksum */
    AG_UINT16_TO_BUFFER(ptUdp->checksum,0);

    /* use dwPreChecksum to calculate pseudo header checksum first */
    dwPreChecksum = ptIp->proto;
    dwPreChecksum += (ptIp->src_ip[0] << 8) | ptIp->src_ip[1];
    dwPreChecksum += (ptIp->src_ip[2] << 8) | ptIp->src_ip[3];

    dwPreChecksum += (ptIp->dst_ip[0] << 8) | ptIp->dst_ip[1];
    dwPreChecksum += (ptIp->dst_ip[2] << 8) | ptIp->dst_ip[3];

    dwPreChecksum += iUdpLen;

    /*check = pseudo header checksum + tcp content checksum */
    wChecksum = CalcChecksum2(dwPreChecksum, (BYTE *)ptUdp, iUdpLen);

    AG_UINT16_TO_BUFFER(ptUdp->checksum,htons(wChecksum));

    return (wChecksum);

}

unsigned long get_br_ip(void)
{
    struct net_device *dev = NULL;
    struct in_device *in_dev = NULL;
    struct in_ifaddr *ifaddr = NULL;
    unsigned long br_ip = DEFAULT_BR_IP;
    unsigned long if_ip;
   
    dev = dev_get_by_name(&init_net,BR_IFNAME);
    if (dev)
    {
        in_dev = in_dev_get(dev);
        dev_put(dev);
    }

    if (in_dev)
    {
        if (in_dev->ifa_list)
        {
            ifaddr = in_dev->ifa_list;
            do
            {
                if_ip = htonl(ifaddr->ifa_local);
                if (if_ip != br_ip)
                {
                    br_ip = if_ip;
                    break;
                }
                ifaddr = ifaddr->ifa_next;
            } while (ifaddr);
        }

        in_dev_put(in_dev);
    }
    return br_ip;
}

int dnsRedirect_construct_reply
(
    struct sk_buff *skb,
    T_IPHeader *pIpHeader, 
    unsigned short queryType
)
{
    struct ethhdr *eth_hdr;
    T_UDPHeader *pUdpHeader;
    dnsHeader_T *dns_header;
    char temp_mac[ETH_ALEN];
    char tmp_ip[4];
    char tmp_port[2];
    char *answer;
    dnsResourceRecord_T dnsResourceRecord;
    unsigned short recordLen = 0;
    unsigned short s_len=0;
    int i;
    unsigned short ip_len, udp_len;

    /* Swap the src and dst MAC */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
    eth_hdr = (struct ethhdr *)(skb_mac_header(skb));
#else
    eth_hdr = (struct ethhdr *)(skb->mac.raw);
#endif
    memcpy(temp_mac, eth_hdr->h_dest, ETH_ALEN);
    memcpy(eth_hdr->h_dest, eth_hdr->h_source, ETH_ALEN);
    memcpy(eth_hdr->h_source, temp_mac, ETH_ALEN);

    /* Swap the IP address */
    memcpy(tmp_ip, pIpHeader->dst_ip, sizeof(tmp_ip));
    memcpy(pIpHeader->dst_ip, pIpHeader->src_ip, sizeof(tmp_ip));
    memcpy(pIpHeader->src_ip, tmp_ip, sizeof(tmp_ip));

    /* Swap the UDP port */
    pUdpHeader = (T_UDPHeader *)GetIpPayload(pIpHeader);
    memcpy(tmp_port, pUdpHeader->src_port, sizeof(tmp_port));
    memcpy(pUdpHeader->src_port, pUdpHeader->dst_port, sizeof(tmp_port));
    memcpy(pUdpHeader->dst_port, tmp_port, sizeof(tmp_port));
    pUdpHeader->checksum[0] = 0;
    pUdpHeader->checksum[1] = 0;

    /* Construct the reply */
    dns_header = (dnsHeader_T *)GetUdpPayload(pIpHeader);
    dns_header->qr = DNS_REPLY;
    dns_header->aa = 0;
    dns_header->ra = 1;
    
    dns_header->numQuestion   = htons(1);
    dns_header->numAnswer     = htons(1);
    dns_header->numAuthority  = htons(0);
    dns_header->numAdditional = htons(0);

    /* Fill in the ANSWER section */
    answer = &(skb->data[skb->len]);

    *answer++ = 0xc0;             /* Use compressed format to represent name */
    *answer++ = sizeof(struct S_dnsHeader);
    skb->len += 2;

    dnsResourceRecord.type  = htons(queryType);     /* Type */
    dnsResourceRecord.class = htons(1);             /* Class : Internet data */
    dnsResourceRecord.ttl   = htonl(0);             /* pling modified 02/07/2007, 3->0, Ntgr spec */

    //skb->dev = get_br_dev();

    if (queryType == DNS_QUERY_TYPE_IP)
    {
        unsigned long ipAddr = get_br_ip();
        dnsResourceRecord.len = htons(4);
        s_len = 4;

        for (i=0; i<4; i++)
        {
            unsigned char byte = (ipAddr >> ((3-i) * 8)) & 0x000000FF;
            dnsResourceRecord.data[i] = byte;
        }
    }
    else
    if (queryType == DNS_QUERY_TYPE_IPV6)
    {
        dnsResourceRecord.len = 0;
        s_len = 0;
    }
    else
    if (queryType == DNS_QUERY_TYPE_PTR)
    {
        unsigned char response[] = "localhost";

        dnsResourceRecord.len    = htons(2 + strlen(response));
        s_len = 2 + strlen(response);
        dnsResourceRecord.data[0] = strlen(response);
        memcpy(&dnsResourceRecord.data[1], response, strlen(response));
        dnsResourceRecord.data[1+strlen(response)] = '\0';
    }
    else
    {
        printf("queryType unknown (%d)\n", queryType);
    }
    recordLen = sizeof(dnsResourceRecord.type) + sizeof(dnsResourceRecord.class) +
                sizeof(dnsResourceRecord.ttl) + sizeof(dnsResourceRecord.len) +
                s_len; //Modified by Erichuang, 6/15/20
    memcpy(answer, &dnsResourceRecord, recordLen);
    skb->len += recordLen;

    /* UDP len and IP len */
    AG_BUFFER_TO_UINT16(udp_len, pUdpHeader->length, htons);
    udp_len += recordLen+2;
    AG_UINT16_TO_BUFFER(pUdpHeader->length, htons(udp_len));

    AG_BUFFER_TO_UINT16(ip_len, pIpHeader->length, htons);
    ip_len += recordLen+2;
    AG_UINT16_TO_BUFFER(pIpHeader->length, htons(ip_len));

    CalcUdpChecksum(pIpHeader);
    CalcIpChecksum(pIpHeader);

    return 0;
}





BOOL dnsRedirect_dnsHookFn
(
    struct sk_buff *skb, 
    T_IPHeader *pIpHeader
)
{
    T_UDPHeader *pUdpHeader;
    unsigned short queryType;

    BR_DNS_HIJACK_DEBUG("dnsRedirect_dnsHookFn: enter>>\n");

    
    if (IsUdpPacket(pIpHeader))
    {
        /*foxconn added start by EricHuang, 02/21/2006*/
        unsigned short fragment;
        memcpy((char *)&fragment, pIpHeader->fragment, 2);
        fragment = ntohs(fragment);
        if (fragment & 0x1fff)  /* ip_off is in host byte order */
            return TRUE;        /*not the first packet*/
        /*foxconn added end by EricHuang, 02/21/2006*/
        
        pUdpHeader=(T_UDPHeader *)GetIpPayload(pIpHeader);
    
        if (UDP_GET_DSTPORT(pUdpHeader) == DNS_PORT)
        {
            BR_DNS_HIJACK_DEBUG("dnsRedirect_dnsHookFn: udp (dns) pkt\n");
            if (dnsRedirect_isNeedRedirect(pIpHeader, &queryType))
            {
                BR_DNS_HIJACK_DEBUG("dnsRedirect_dnsHookFn: sendreply\n");
                dnsRedirect_construct_reply(skb, pIpHeader, queryType);
                return FALSE;
            }
            
            return TRUE;
        }
    } /* end if (IsUdpPacket(pIpHeader)) */

    return TRUE;
}

static BOOL dnsFilterHook(struct sk_buff *skb)
{
    T_IPHeader *ptIp;

    BR_DNS_HIJACK_DEBUG("dnsFilterHook: enter>>\n");

    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)    
    ptIp = (T_IPHeader*)skb_network_header(skb);  
#else
    ptIp = (T_IPHeader*)(skb->nh.iph);
#endif
    

    if (!dnsRedirect_dnsHookFn(skb, ptIp))
    {
        return TRUE;
    }

    return FALSE;
}

static int br_dns_hijack_hook
(
    struct sk_buff *skb,
    unsigned int hooknum,
    const struct net_device *in_dev,
    const struct net_device *out_dev
)
{
    BOOL bResult = BR_ACCEPT;
    
   
    BR_DNS_HIJACK_DEBUG("before br_dns_hijack_hook : enter>>\n");
   
    struct net_device *br0_dev;
    struct in_device *br0_in_dev;

   
    /* Only hijack packets from LAN / WLAN (AP) interfaces */
    if (strcmp(in_dev->name, WLAN_IFNAME) == 0 ||
        strcmp(in_dev->name, WLAN_GUEST_IFNAME) == 0 ||
        strcmp(in_dev->name, LAN_IFNAME) == 0)
    {
        BR_DNS_HIJACK_DEBUG("br_dns_hijack_hook: enter>>\n");
        if( dnsFilterHook(skb) == TRUE)
        {
            /* Modify the output device to send the packet
             * back to original interface.
             */
            skb->dev = (struct net_device *)in_dev;

            /* For local in, don't return to bridge since
             * bridge will call 'netif_rx' after the local_in hook.
             * We sent the skb to back to the incoming device directly.
             */
            if (hooknum == NF_BR_LOCAL_IN)
            {
                skb_push(skb, ETH_HLEN);
                dev_queue_xmit(skb);
                return BR_DROP;
            }
        }    
    }

    return bResult;
}

static unsigned int br_preroute_hook
(
    unsigned int hook,
    struct sk_buff *skb,
    const struct net_device *in,
    const struct net_device *out,
    int (*okfn)(struct sk_buff *)
)
{
    BR_DNS_HIJACK_DEBUG("%s: hook=%d in_dev='%s', out_dev='%s'\n",
            __FUNCTION__, hook, in->name, out->name);
    switch (br_dns_hijack_hook(skb, hook, in, out))
    {
        case BR_ACCEPT:
            return NF_ACCEPT;
        case BR_DROP:
        case BR_QUEUE:
            return NF_STOLEN;
    }
    return NF_DROP;
}

static struct nf_hook_ops br_forward_ops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
    .owner          = THIS_MODULE,
#else
    .list           = {NULL, NULL},
#endif
    .hook           = br_preroute_hook,
    .pf             = PF_BRIDGE,
    .hooknum        = NF_BR_FORWARD,
    .priority       = INT_MIN,
};

static unsigned int br_local_in_hook
(
    unsigned int hook,
    struct sk_buff *skb,
    const struct net_device *in,
    const struct net_device *out,
    int (*okfn)(struct sk_buff *)
)
{
    BR_DNS_HIJACK_DEBUG("%s: hook=%d in_dev='%s', out_dev='%s'\n",
            __FUNCTION__, hook, in->name, out->name);
    switch (br_dns_hijack_hook(skb, hook, in, out))
    {
        case BR_ACCEPT:
            return NF_ACCEPT;
        case BR_DROP:
        case BR_QUEUE:
            return NF_STOLEN;
    }
    return NF_DROP;
}

static struct nf_hook_ops br_local_in_ops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
    .owner          = THIS_MODULE,
#else
    .list           = {NULL, NULL},
#endif
    .hook           = br_local_in_hook,
    .pf             = PF_BRIDGE,
    .hooknum        = NF_BR_LOCAL_IN,
    .priority       = INT_MIN,
};

static int br_dns_hijack_init(void)
{


    if (nf_register_hook(&br_forward_ops) < 0)
        printk(KERN_EMERG "Unable to register NF_BR_FORWARD hook!\n");
    if (nf_register_hook(&br_local_in_ops) < 0)
        printk(KERN_EMERG "Unable to register NF_BR_LOCAL_IN hook!\n");

    /*added by dennis, foxconn, netlink communication*/
    netlink_sock = netlink_kernel_create(&init_net, NETLINK_USERSOCK, 0,udp_receive, NULL, THIS_MODULE);
    /* readyshare name */
    memset(readyshare_name, 0, sizeof(readyshare_name));
    strcpy(readyshare_name, readyshare_dev);

    return 0;
}



static void br_dns_hijack_exit(void)
{
    nf_unregister_hook(&br_forward_ops);
    nf_unregister_hook(&br_local_in_ops);
    sock_release(netlink_sock->sk_socket);
    printk("netlink driver remove successfully\n");
}

module_init(br_dns_hijack_init);
module_exit(br_dns_hijack_exit);
