/***************************************************************************
#***
#***    Copyright 2005  Hon Hai Precision Ind. Co. Ltd.
#***    All Rights Reserved.
#***    No portions of this material shall be reproduced in any form without the
#***    written permission of Hon Hai Precision Ind. Co. Ltd.
#***
#***    All information contained in this document is Hon Hai Precision Ind.  
#***    Co. Ltd. company private, proprietary, and trade secret property and 
#***    are protected by international intellectual property laws and treaties.
#***
#****************************************************************************/

#ifndef _acosnvram_h
#define _acosnvram_h

#include <acosTypes.h>


#define PATH_DEV_NVRAM "/etc/parameter"
#define NVRAM_CFG_FILENAME  "/fl/NVRAM"
/* wklin modified start */
#define NVRAM_MAX_PARAM_LEN (4*1024) /* max len for a sigle parameter */
/* wklin modified end */
struct nvram_header {
	unsigned long magic;
	unsigned long len;
        unsigned short unused;
	unsigned short checksum;
	unsigned long version;
	unsigned long reserved[4];
};

struct nvram_tuple {
	char *name;
	char *value;
	struct nvram_tuple *next;
};

/* Compatibility */
typedef struct nvram_tuple EnvRec;


extern int nvram_loaddefault(void);

/*
 * Initialize NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern int nvram_init(void);

/*
 * Disable NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern void nvram_exit(void);

/*
 * Get the value of an NVRAM variable. The pointer returned may be
 * invalid after a set.
 * @param	name	name of variable to get
 * @return	value of variable or NULL if undefined
 */
/* wklin modified start */
extern char * nvram_get(const char *name, char *buf, int buf_len);
/* wklin modified end */
/* 
 * Get the value of an NVRAM variable.
 * @param	name	name of variable to get
 * @return	value of variable or NUL if undefined
 */
/* wklin modified start */
#define nvram_safe_get(name, buf, len) (nvram_get(name, buf, len) ? : "")
/* wklin modified end */
/*
 * Match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is string equal
 *		to match or FALSE otherwise
 */
static INLINE int
nvram_match(char *name, char *match) {
    /* wklin modified start */
    char tmp[NVRAM_MAX_PARAM_LEN];
	const char *value = nvram_get(name,tmp,NVRAM_MAX_PARAM_LEN);
    /* wklin modified end */
	return (value && !strcmp(value, match));
}

/*
 * Inversely match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is not string
 *		equal to invmatch or FALSE otherwise
 */
static INLINE int
nvram_invmatch(char *name, char *invmatch) {
    /* wklin modified start */
    char tmp[NVRAM_MAX_PARAM_LEN];
	const char *value = nvram_get(name, tmp, NVRAM_MAX_PARAM_LEN);
    /* wklin modified end */
	return (value && strcmp(value, invmatch));
}

/*
 * Set the value of an NVRAM variable. The name and value strings are
 * copied into private storage. Pointers to previously set values
 * may become invalid. The new value may be immediately
 * retrieved but will not be permanently stored until a commit.
 * @param	name	name of variable to set
 * @param	value	value of variable
 * @return	0 on success and errno on failure
 */
extern int nvram_set(const char *name, const char *value);

/*
 * Unset an NVRAM variable. Pointers to previously set values
 * remain valid until a set.
 * @param	name	name of variable to unset
 * @return	0 on success and errno on failure
 * NOTE: use nvram_commit to commit this change to flash.
 */
extern int nvram_unset(const char *name);

/*
 * Commit NVRAM variables to permanent storage. All pointers to values
 * may be invalid after a commit.
 * NVRAM values are undefined after a commit.
 * @return	0 on success and errno on failure
 */
extern int nvram_commit(void);

/*
 * Get all NVRAM variables (format name=value\0 ... \0\0).
 * @param	buf	buffer to store variables
 * @param	count	size of buffer in bytes
 * @return	0 on success and errno on failure
 */
extern int nvram_getall(char *buf, int count);

/* Foxconn added start Peter Ling 12/05/2005 */
/* used for config backup and restore */
extern int nvram_readflash(char *buf);
extern int nvram_writeflash(char *buf, int size);
/* Foxconn added end Peter Ling 12/05/2005 */

#define NVRAM_MAGIC		0x40112791
#define NVRAM_VERSION		1
#define NVRAM_HEADER_SIZE	32
#ifdef NVRAM_SIZE_16K
#define NVRAM_SPACE		0x4000
#else
#ifdef U12H139
#define NVRAM_SPACE		0x6000
#else
#define NVRAM_SPACE		0x10000
#endif
#endif
/* Foxconn added start , Jasmine Yang, 12/12/2005 */
extern unsigned short hd_cksum(unsigned short *addr, int len);
/* Foxconn added end , Jasmine Yang, 12/12/2005 */

#endif /* _acosnvram_h */
